require 'resolv'
Puppet::Functions.create_function(:'netinfo::get_prefix') do
	dispatch :get_prefix do
		param 'String', :name
	end

	# Get prefix for given name.
	# @param [String] name to be resolved
	# @return [Hash] Hash {ipv4=>[...],ipv6=>[...]} of all addresses found in the first step that matched, nil if no match is found
	#
	# If the name is an IPv4 or IPv6 address, it is returned without further processing.
	# Resolving is done by:
	# - Lookup in network::networks
	# - Lookup in network::external_networks
	# - Lookup in netinfo::dns::forward (name/address mapping from node definitions)
	# - DNS lookup
	# The first match is returned, i.e. if the name exists in network::networks, lookup in network::external_networks etc.
	# is not tried.
	def get_prefix(name)
		name=name.downcase
		#check if this is an IPv4 address
		if(/^([012]?[0-9]{1,2}\.){3}[012]?[0-9]{1,2}(\/([0-2]?[0-9]|3[0-2]))?$/=~name)
			return {'ipv4'=>[name],'ipv6'=>[]}
		end

		#check if this is an IPv6 prefix
		if(call_function('netinfo::validate_ipv6',name))
			return {'ipv4'=>[],'ipv6'=>[name]}
		end

		#split host and domain if given
		if(!(/(?<host>[^\.]*)\.(?<domain>.*)/=~name))
			host=name
			domain=call_function("lookup","dns::primary_zone")
		elsif (name=~/localhost/i)
			return {'ipv4'=>['127.0.0.1/128'],'ipv6'=>['::1/128']}
		end

		#initialize result
		res={'ipv4'=>[],'ipv6'=>[]}

		#check network::networks
		nets=downcase_keys(call_function("lookup","network::networks"))
		if(nets.key?(name))
			addr=nets[name]
			if(/^([012]?[0-9]{1,2}\.){3}[012]?[0-9]{1,2}(\/[0-2]?[0-9])?$/=~addr)
				res['ipv4']=[addr]
			elsif (call_function('netinfo::validate_ipv6',addr))
				res['ipv6']=[call_function("netinfo::reduce_ipv6",addr)]
			else
				fail("netinfo::get_prefix: entry #{name}: #{addr} is neither an IPv4 nor an IPv6 address")
			end
			return res
		end

		#check network::external_networks
        nets=downcase_keys(call_function("lookup","network::external_networks",Puppet::Pops::Types::PHashType::DEFAULT,"hash",{}))
        if net=downcase_keys(nets).dig(name)
		found=false
          if(net.key?('ipv4'))
		    found=true
            r=net['ipv4']
            if(r.is_a?(Array))
              res['ipv4']=r
            else
              res['ipv4']=[r]
            end
		  end
          if(net.key?('ipv6'))
		    found=true
            r=net['ipv6']
            if(r.is_a?(Array))
              res['ipv6']=r
            else
              res['ipv6']=[r]
            end
		  end
          if(found)
            return res
          end
		end

		#initialize resolver, may be used from get_from_local_dns if address is a CNAME to an external host
		@resolver=Resolv::DNS.new()
		#check own DNS
		@dns=call_function("lookup","netinfo::dns::forward")
		puts "local dns for #{host}.#{domain}"
		res=get_from_local_dns(host,domain)
		if(res)
			res.keys.each do|k|
				len=k=~/ipv4/i?'32':'128'
				res[k]=res[k].map{|v| "#{call_function('netinfo::reduce_ipv6',v.downcase)}"}
			end
			return res
		end
		#fall back to global DNS lookup
		puts "global dns for #{host}.#{domain}"
		get_from_global_dns(name)
	end

	def get_from_local_dns(host,domain)
	puts "#{host}.#{domain}"
		res={'ipv4'=>[],'ipv6'=>[]}
		found=false
		@dns=call_function("lookup","netinfo::dns::forward")
		if(@dns.key?(domain))
			if(@dns[domain].key?('AAAA'))
				if(@dns[domain]['AAAA'].key?(host))
					res['ipv6']=@dns[domain]['AAAA'][host].map{|s| s.downcase}
					found=true
				end
			end
			if(@dns[domain].key?('A'))
				if(@dns[domain]['A'].key?(host))
					res['ipv4']=@dns[domain]['A'][host].map{|s| s.downcase}
					found=true
				end
			end
			if(found)
				return res
			end
			if(@dns[domain].key?('CNAME'))
				if(@dns[domain]['CNAME'].key?(host))
					name=@dns[domain]['CNAME'][host]
					t_domain=domain
					puts "CNAME #{host} -> #{name}"
					if(!(/(?<host>[^\.]*)\.(?<domain>.*[^\.])\.?$/=~name))
						puts "CNAME #{name} matched regex"
						host=name
						domain=t_domain
					end
					puts "CNAME get_local_dns(#{host},#{domain})"
					r=get_from_local_dns(host,domain)
					if(r)
						return r
					else
						r=get_from_global_dns(host+"."+domain)
						if(r)
							return r
						end
					end
				end
			end
		end
		nil #if something was found, this is not reached
	end

	# look up a name in global DNS
	# @param [String] hostname to look up
	# @return [Hash] Hash {ipv4=>[...],ipv6=>[...]} of all addresses found in the first step that matched, nil if no match is found
	#
	# If at least one A or AAAA record exists, all IP addresses references by the A and AAAA records are returned. If the name resolves
	# to a CNAME, get_from_global_dns is called recursively until either a A/AAAA record is found or resolving fails.
	def get_from_global_dns(hostname)
		res={'ipv4'=>[],'ipv6'=>[]}
		found=false
		@resolver.getresources(hostname,Resolv::DNS::Resource::IN::A).each do|a|
			res['ipv4'] << "#{a.address.to_s}/32"
			found=true
		end
		@resolver.getresources(hostname,Resolv::DNS::Resource::IN::AAAA).each do|a|
			res['ipv6'] << "#{a.address.to_s}/128"
			found=true
		end
		if(found)
			return res
		else
			@resolver.getresources(hostname,Resolv::DNS::Resource::IN::CNAME).each do|a|
				r=get_from_global_dns(a.name)
				if(r)
					res['ipv4']+=r['ipv4']
					res['ipv6']+=r['ipv6']
					res['ipv4'].uniq!
					res['ipv6'].uniq!
					return(res)
				end
			end
		end
		nil
	end

	def downcase_keys(h)
		r={}
		h.each do|k,v|
			key=k.downcase
			if(r.key?(key))
				fail("netinfo: key #{key} is not unique when downcased!")
			end
			r[key]=v
		end
		r
	end
end
