# The `yaml_cluster` is a hiera 5 `data_hash` data provider function.
# See [the configuration guide documentation](https://puppet.com/docs/puppet/latest/hiera_config_yaml_5.html#configuring-a-hierarchy-level-built-in-backends) for
# how to use this function.
#
# @since 4.8.0
#
require 'yaml'
require 'digest/md5'
require 'socket'

Puppet::Functions.create_function(:yaml_netinfo) do
    dispatch :yaml_netinfo do
        param 'Hash', :options
        param 'Puppet::LookupContext', :context
    end

    argument_mismatch :missing_path do
        param 'Hash', :options
        param 'Puppet::LookupContext', :context
    end

    def get_yaml(path,context)
        context.cached_file_data(path) do |content|
        begin #this block is run as lambda in cached_file_data on file content
            data = YAML.load(content, path)
            if data.is_a?(Hash)
                data
            else
                msg = _("%{path}: file does not contain a valid yaml hash" % { path: path })
                raise Puppet::DataBinding::LookupError, msg if Puppet[:strict] == :error && data != false
                Puppet.warning(msg)
            end
            rescue YAML::SyntaxError => ex
                # Psych errors includes the absolute path to the file, so no need to add that
                # to the message
                raise Puppet::DataBinding::LookupError, "Unable to parse #{ex.message}"
            end
        end
    end

    def yaml_netinfo(options, context)
    #TODO: implement explicitly defined zones
      basepath=Puppet[:codedir]+"/environments/"+closure_scope['environment']+"/data"
      scope=closure_scope
      net_objects={}
      primary_zone=call_function("lookup","dns::primary_zone")
      ptr_zone=call_function("lookup","dns::ptr_zone")
      nat64_ipv4_prefix=call_function("lookup","network::nat64::ipv4_prefix",Puppet::Pops::Types::PStringType::DEFAULT,'first','10.0.0.0/8')
      if(/^((1?[0-9]{1,2}|2([0-4][0-9]|5[0-5]))\.){3}(1?[0-9]{1,2}|2([0-4][0-9]|5[0-5]))\/([12]?[0-9]|3[0-2])$/=~nat64_ipv4_prefix)
        s=nat64_ipv4_prefix.split('/')
        nat64_ipv4_blocks=4-s[1].to_i/8 #number of blocks that are available for mapping
        if(s[1].to_i%8!=0)
          fail("netinfo: network::nat64::ipv4_prefix = '#{nat64_ipv4_prefix}' contains invalid prefix length, only length dividable by 8 are supported.")
        end
        nat64_ipv4_prefix=s[0]
        s=nat64_ipv4_prefix.split('.')
        nat64_ipv4_prefix=''
        #copy blocks that are not avaliable for mapping
        for i in 0..3-nat64_ipv4_blocks
            nat64_ipv4_prefix+="#{s[i]}."
        end
        nat64_ipv4_prefix=nat64_ipv4_prefix[0,nat64_ipv4_prefix.size-1]
      else
        fail("netinfo: network::nat64::ipv4_prefix #{nat64_ipv4_prefix} is invalid.")
      end
      nat64_ipv4_prefix_full=nat64_ipv4_prefix # IPv4 prefix in CIDR notation
      nat64_ipv4_router_address=nat64_ipv4_prefix # address of the router itself
      for  i in 1..nat64_ipv4_blocks
        nat64_ipv4_prefix_full+=".0"
        nat64_ipv4_router_address+=".255"
      end
      nat64_ipv4_router_address[nat64_ipv4_router_address.size-1]='4' #change last 255 to 254 (as 255 is reserved)
      nat64_ipv4_prefix_full+="/"+((4-nat64_ipv4_blocks)*8).to_s
      nat64_mapping={}
      records={primary_zone=>{'AAAA'=>{},'CNAME'=>{}}}
      ptr={ptr_zone=>{'PTR'=>{}}}
      node_primary_address=''
      ns=[]
      db_servers={}
      dns_servers={}
#      default_config=call_function("lookup","dns::zone_default_config",Puppet::Pops::Types::PHashType::DEFAULT,"hash",{})
#      #config for primary zone
#      configured_zones=call_function("lookup","dns::zones",Puppet::Pops::Types::PHashType::DEFAULT,"hash",{})
#      if(configured_zones.has_key?(primary_zone))
#        records[primary_zone]['config']=configured_zones[primary_zone]
#      else
#        if(default_config.nil? || default_config.empty?)
#          fail("primary zone #{primary_zone} has no explicit config and dns::zone_default_config is undefined!")
#        end
#        records[primary_zone]['config']=default_config
#      end
#      #config for ptr zone
#      configured_zones=call_function("lookup","dns::zones",Puppet::Pops::Types::PHashType::DEFAULT,"hash",{})
#      if(configured_zones.has_key?(ptr_zone))
#        ptr['config']=configured_zones[ptr_zone]
#      else
#        if(default_config.nil? || default_config.empty?)
#          fail("PTR zone #{ptr_zone} has no explicit config and dns::zone_default_config is undefined!")
#        end
#        ptr['config']=default_config
#      end
      #generate primary and ptr zone
      #iterate over all nodes
      Dir.entries("#{basepath}/#{options['node_path']}").each do |file|
        if(("."==file)||(".."==file)||("template.yaml"==file)||(/.yaml$/!~file))
          next
        end
        node=file.gsub(/.yaml$/,'')
        if(/^adm-dns/=~node)
          ns<<node+"."+primary_zone
        end
        path="#{basepath}/#{options['node_path']}/#{file}"
        data=get_yaml(path,context)
        if(!data.is_a?(Hash))
          next
        end
        res=get_vm_dns_records(node,data,primary_zone)
        if(!res['AAAA'].empty?)
          node_primary_address=res['AAAA'][node][0]
          records[primary_zone]['AAAA'].merge!(res['AAAA'])
          if((/[a-z]{2,3}-itf-\w{6}-app-imex-01/=~node) || (/[a-z]{2,3}-[a-z]{3}-([a-z]{4,7}-|)dbs-[0-9]{2}/=~node) || (/adm-dbs-[0-9]{2}/=~node))
            db_servers.merge!({node=>res['AAAA'][node][0]})
          end
          if /^adm-dns-\d{2}/=~node
            dns_servers.merge!({node=>res['AAAA'][node][0]})
          end
          if(!res['PTR'].empty?)
            ptr[ptr_zone]['PTR'].merge!(res['PTR'])
          end
          if(!res['CNAME'].empty?)
            records[primary_zone]['CNAME'].merge!(res['CNAME'])
          end
          if(data.key?('network::nat64_id'))
            if(!node_primary_address.empty?)
              mapped_ipv4=get_ipv4_mapping_address(data['network::nat64_id'],nat64_ipv4_prefix,nat64_ipv4_blocks)
              if(mapped_ipv4)
                nat64_mapping.merge!(node=>{mapped_ipv4=>node_primary_address})
              end
            end
          end
        end
      end
      #add net_objects from A and AAAA records
      records.each do |zone,data|
        data.each do |type,entries|
          entries.each do |node,addrs|
            if (type=='AAAA' or type='A') and !net_objects.key?(node)
              net_objects[node+"."+zone]={'ipv4'=>[],'ipv6'=>[]}
              if zone==primary_zone
                net_objects[node]={'ipv4'=>[],'ipv6'=>[]}
              end
            end
            if type=='AAAA'
              ipver='ipv6'
            elsif type=='A'
              ipver='ipv4'
            end
            net_objects[node+"."+zone][ipver]=addrs
            if zone==primary_zone
              net_objects[node][ipver]=addrs
            end
          end
        end
      end
      #add net_objects from CNAMES
      records.each do |zone,data|
        if data.key?('CNAME')
          entries=data['CNAME']
          entries.each do |node,cname|
            if net_objects.key?(cname)
              #address pointed to is avaliable in net_objects, copy it
              net_objects[node]=net_objects[cname]
            else
              # address pointed to is not available, do DNS lookup
              # TODO
            end
          end
        end
      end
      #add networks
      net_prefix=call_function('lookup','network::net-prefix',Puppet::Pops::Types::PStringType::DEFAULT,"first",'')
      networks=call_function('lookup','network::networks',Puppet::Pops::Types::PHashType::DEFAULT,"hash",{})
      networks.keys.each do|k|
        key=k.downcase.gsub(/^#{net_prefix}/,'')
        net_objects[key]=call_function('netinfo::get_prefix',networks[k])
      end
      #get views
      dns64_view_overrides=get_dns64_views_and_overrides()
      #get dns64 ignore nets
      dns_acls=get_dns_acls(call_function('lookup','network::nat64::exclude_nets',Puppet::Pops::Types::PArrayType::DEFAULT,'unique',[]))
      res={
        "netinfo::dns::NS"=>ns,
        "netinfo::dns::forward"=>records,
        "netinfo::dns::reverse"=>ptr,
        "netinfo::dns::acls"=>dns_acls,
        "netinfo::db_servers"=>db_servers,
        "netinfo::dns_servers"=>dns_servers,
        "netinfo::nat64::mapping"=>nat64_mapping,
        "netinfo::nat64::ipv4_prefix"=>nat64_ipv4_prefix_full,
        "netinfo::nat64::ipv4_router_address"=>nat64_ipv4_router_address,
        "netinfo::dns::dns64_views"=>dns64_view_overrides['views'],
        "netinfo::dns::dns64_view_overrides"=>dns64_view_overrides['overrides'],
        "netinfo::net_objects"=>net_objects,
        "netinfo::nat64::all_ipv6_prefixes"=>get_all_ipv6_nat_prefixes
      }
      if scope.exist?('hs') && scope.exist?('opmode')
        res.merge!("netinfo::nat64::ipv6_prefix"=>get_ipv6_nat_prefix(scope['hs_net'],scope['opmode']))
      end
      res
    end

    def get_dns_acls(nat64_excluded_nets)
      res={}
      unless nat64_excluded_nets.empty?
        res['nat64_exclude_nets']=nat64_excluded_nets.map{|k| "!#{k}"}+['any']
      end
      res
    end
    def get_all_ipv6_nat_prefixes
        prefix_tree=call_function('lookup','network::nat64::prefix_tree',Puppet::Pops::Types::PHashType::DEFAULT,"hash",{})
        if(!prefix_tree.key?('customers'))
            fail("network::nat64::prefix_tree does not contain an customer hash")
        end
        if(!prefix_tree.key?('opmode'))
            fail("network::nat64::prefix_tree does not contain an opmode hash")
        end
        prefix_tree['customers'].map{|customer| prefix_tree['opmode'].map{|opmode| get_ipv6_nat_prefix(customer[0],opmode[0])}}.flatten
    end
    # get ipv6 nat prefix:
    # - base network is retrieved from network::nat64::prefix_tree/base_network which is looked up in network::networks
    # - if network::nat64::prefix_tree/padding/pad_after_id is false:
    #   - this is filled up to /80 with network::nat64::prefix_tree/padding/pad_char (default to '0') , i.e. base network must be /80 or larger
    #   - one byte for customer and one byte for opmode is appended, reachting /96
    # - else
    #   - one byte for customer and one byte for opmode is appended to the base prefix
    #   - the remaining adress is filled with pad_char (c.f. above) up to /96
    
    def get_ipv6_nat_prefix(customer,opmode)
        prefix_tree=call_function('lookup','network::nat64::prefix_tree',Puppet::Pops::Types::PHashType::DEFAULT,"hash",{})
        prefix_tree['base_network'].downcase!
        net_prefix=call_function('lookup','network::net-prefix',Puppet::Pops::Types::PStringType::DEFAULT,"first",'')
        networks=call_function('lookup','network::networks',Puppet::Pops::Types::PHashType::DEFAULT,"hash",{})
        networks.keys.each do|k|
            key=k.downcase.gsub(/^#{net_prefix}/,'')
            networks[key]=networks[k]
            networks.delete(k)
        end
        if(!prefix_tree.key?('base_network'))
            fail("network::nat64::prefix_tree does not contain a base_network entry")
        end
        if(!prefix_tree.key?('customers'))
            fail("network::nat64::prefix_tree does not contain an customer hash")
        end
        if(!prefix_tree['customers'].key?(customer))
            fail("network::nat64::prefix_tree does not contain an entry for customer '#{customer}'")
        end
        if(!networks.key?(prefix_tree['base_network']))
            fail("network::networks does not contain a network #{prefix_tree['base_network']} referenced by network::nat64::prefix_tree/base_network")
        end
        if prefix_tree.key?('padding')
          pad_with=prefix_tree['padding'].key?('pad_char')?prefix_tree['padding']['pad_char']:'0'
          pad_after_identifier=prefix_tree['padding'].key?('pad_after_id')?prefix_tree['padding']['pad_after_id']:true
        else
          pad_with='0'
          pad_after_identifier=true
        end
        n=networks[prefix_tree['base_network']].split('/')
        if(n.size!=2)
            fail("network::networks/#{prefix_tree['base_network']} has invalid format, must contain prefix and prefix length")
        end
        net=n[0]
        len=n[1].to_i
        if(len>80)
            fail("network #{prefix_tree['base_network']} referenced by network::nat64::prefix_tree/base_network must be a /80 or larger")
        end
        if(len%4!=0)
            fail("network #{prefix_tree['base_network']} referenced by network::nat64::prefix_tree/base_network must have a prefix length dividable by 4")
        end    
        skip_blocks=len/4
        ext_prefix=call_function('netinfo::extend_ipv6',networks[prefix_tree['base_network']])
        t=ext_prefix.split('/')
        p=t[0]
        skip_blocks=t[1].to_i/4
        res=p[0,4] #first two byte
        #copy address up to original prefix length
        offset=0
        for i in 4..skip_blocks-1
          if(i%4==0)
            res+=":"
            offset+=1
          end
          res+=p[i+offset]
        end
        if(pad_after_identifier==true)
          cust_opmode_separator=""
          case skip_blocks%4
            when 0
              #res is generated up to end of block, add :
              res+=":"
            when 3
              #res is one byte before end of block -> place : between customer and opmode
              cust_opmode_separator=":"
          end
          #set customer/opmode bytes
          res+=sprintf('%01x%s%01x',prefix_tree['customers'][customer],cust_opmode_separator,prefix_tree['opmode'][opmode]);
          skip_blocks+=2
          #fill up to /96 with pad_with
          for i in skip_blocks..23
            if(i%4==0)
              res+=":"
            end
            res+=pad_with
          end
        else
          #fill up to /80 with all-ones
          for i in skip_blocks..21
            if(i%4==0)
              res+=":"
            end
            res+=pad_with
          end
          #set customer/opmode bytes
          res+=sprintf('%01x%01x',prefix_tree['customers'][customer],prefix_tree['opmode'][opmode])
        end
        res+="::/96"

        call_function("netinfo::reduce_ipv6",res)
    end

    def get_ipv4_mapping_address(id,net,len)
        id=id.to_i
        len=len.to_i
        addr="."+(id%255).to_s
        id/=255
        i=1;
        while(id>0)
            addr="."+(id%256).to_s+addr
            id/=256
            i+=1
        end
        if(i>len)
            return(nil)
        else
          for j in i..len-1
            addr=".0"+addr
          end
        end
        return(net+addr)
    end

    def get_dns64_views_and_overrides
        views={} #nat_prefix => {net1=>net1_prefix,net2=>net2_prefix,...}
        overrides={} #nat_prefix=>{AAAA=>{host=>addr,...},CNAMES=>{....},A=>{...},SRV=>{...},TXT=>{...},drop=>[...]}
        proxy_map=call_function('lookup','network::nat64::proxy_mapping',Puppet::Pops::Types::PHashType::DEFAULT,"hash",{})
        override_spec=call_function('lookup','dns::overrides',Puppet::Pops::Types::PHashType::DEFAULT,"hash",{})
        #sort proxy may by priority to proxies
        proxies={}
        i=0
        while (proxy_map.size>0)
            max_prio=-1;
            current=''
            proxy_map.each do|name,opts|
              unless opts.key?('priority')
                fail("network::nat64::proxy_mapping/#{name}: priority missing")
              end
                if(opts['priority']>max_prio)
                    max_prio=opts['priority']
                    current=name
                end
            end
            proxies[i]={'name'=>current,'opts'=>proxy_map[current]}
            i=i+1
            proxy_map.delete(current)
        end
        
        #iterate over all networks and match against proxies
        nets=call_function('lookup','network::networks',Puppet::Pops::Types::PHashType::DEFAULT,"hash",{})
        nets.each do|net,prefix|
            proxies.each do|num,proxy|
              rulename=proxy['name']
              opts=proxy['opts']
                if(captures=/#{opts['pattern']}/.match(net))
					if(opts.key?('disable') and opts['disable'])
					  nat_prefix="disable"
					  nat_prefix_netpart=""
					else
						customer=replace_capture_marks(opts['customer'],captures.to_a)
						opmode=replace_capture_marks(opts['opmode'],captures.to_a)
						nat_prefix=get_ipv6_nat_prefix(customer,opmode)
						nat_prefix_netpart=nat_prefix.gsub(/\/[0-9]+/,'')
					end
                    if(views.key?(nat_prefix))
                        views[nat_prefix].merge!({net=>prefix})
                    else
                        views[nat_prefix]={net=>prefix}
                    end
                    #rule matched, check if override_specs apply
                    override_spec.each do |spec,config|
                      if config.key?('proxy_mappings') and config['proxy_mappings'].include?(rulename)
                        if config.key?('customer')
                          override_customer=config['customer']
                          unless override_customer.is_a?(Array)
                            override_customer=[override_customer]
                          end
                          unless override_customer.include?(customer)
                            next
                          end
                        end
                        keys=['CNAME','SRV','TXT','rpz_options']
                        unless overrides.key?(nat_prefix)
                          overrides[nat_prefix]={
                            'A'=>{},
                            'AAAA'=>{},
                            'CNAME'=>{},
                            'SRV'=>{},
                            'TXT'=>{},
                            'rpz_options' => {}
                          }
                        end
                        keys.each do |key|
                          if(config.key?(key))
                            overrides[nat_prefix][key].merge!(config[key])
                          end
                        end
                        if config.key?('A')
                          config['A'].each do |name,addr|
                            if addr.is_a?(Array)
                              overrides[nat_prefix]['A'].merge!(name=>addr)
                            elsif addr.is_a?(String)
                              overrides[nat_prefix]['A'].merge!(name=>[addr])
                            else
                              fail("dns::overrides/#{rulename}/A/#{name}: invalid address type, must be string or array of strings.")
                            end
                          end
                        end
                        if config.key?('AAAA')
                          config['AAAA'].each do |name,addr|
                            if addr.is_a?(Array)
                              overrides[nat_prefix]['AAAA'].merge!(name=>addr.map{|a| call_function('netinfo::reduce_ipv6',a.gsub(/__NAT64__/,nat_prefix_netpart).gsub(':::','::')).gsub(/\/[0-9]+/,'')})
                            elsif addr.is_a?(String)
                              overrides[nat_prefix]['AAAA'].merge!(name=>[call_function('netinfo::reduce_ipv6',addr.gsub(/__NAT64__/,nat_prefix_netpart).gsub(':::','::')).gsub(/\/[0-9]+/,'')])
                            else
                              fail("dns::overrides/#{rulename}/AAAA/#{name}: invalid address type, must be string or array of strings.")
                            end
                          end
                        end
                        if config.key?('drop')
                          config['drop'].each do |drop|
                            overrides[nat_prefix]['CNAME'].merge!({drop=>'.'})
                          end
                        end
                      end
                    end
                    break
                end
            end
        end

        {'views'=>views,'overrides'=>overrides}
    end
    #TODO: remove get_dns_views when get_dns64_views_overrides is ok
    def get_dns64_views
        res={} #nat_prefix => {net1=>net1_prefix,net2=>net2_prefix,...}
        proxy_map=call_function('lookup','network::nat64::proxy_mapping',Puppet::Pops::Types::PHashType::DEFAULT,"hash",{})
        #sort proxy may by priority to proxies
        proxies=[]
        while (proxy_map.size>0)
            max_prio=-1;
            current=''
            proxy_map.each do|name,opts|
                if(opts['priority']>max_prio)
                    max_prio=opts['priority']
                    current=name
                end
            end
            proxies << proxy_map[current]
            proxy_map.delete(current)
        end
        
        #iterate over all networks and match against proxies
        nets=call_function('lookup','network::networks',Puppet::Pops::Types::PHashType::DEFAULT,"hash",{})
        nets.each do|net,prefix|
            proxies.each do|opts|
                if(captures=/#{opts['pattern']}/.match(net))
                    customer=replace_capture_marks(opts['customer'],captures.to_a)
                    opmode=replace_capture_marks(opts['opmode'],captures.to_a)
                    nat_prefix=get_ipv6_nat_prefix(customer,opmode)
                    if(res.key?(nat_prefix))
                        res[nat_prefix].merge!({net=>prefix})
                    else
                        res[nat_prefix]={net=>prefix}
                    end
                    break
                end
            end
        end
        res
    end
    
    def replace_capture_marks(s,captures)
        last_backslash=false
        last_dollar=false
        current_mark=''
        res=''
        last_replace_at=0
        count_since_last_replace=-1
        for i in 0..s.size
            if(last_dollar && /\d/=~s[i]) #number following non-escaped $ -> capture mark
                current_mark=s[i]
                last_dollar=false
            elsif(!current_mark.empty?)
                if(/\d/=~s[i]) #reading mark, found another number
                    current_mark+=s[i]
                else #reading mark, no number -> end of mark
                    if(count_since_last_replace>0)
                        res+=s[last_replace_at,count_since_last_replace-current_mark.size]
                        count_since_last_replace=-1
                        last_replace_at=i
                    end
                    if(captures.size>current_mark.to_i)
                        res+=captures[current_mark.to_i].to_s
                    end
                    current_mark=''
                end
            else
            end
            if (s[i]=='\\')
                last_backslash=!last_backslash #'\\' is an escape for '\', i.e. ignore even number of backslashes
                last_dollar=false
            elsif (s[i]=='$')
                if(!last_backslash)
                    last_dollar=true
                    last_backslash=false
                else
                    last_dollar=false
                end 
            end    
            count_since_last_replace+=1
            i+=1 
        end     
        res+=s[last_replace_at,s.size]
    end

    def get_dns_record(type,host,addr)
        sprintf("%-40s%-10s%s\n",host.downcase,type.upcase,addr.downcase)
    end

    def get_vm_dns_records(vmname,vmconfig,primary_zone)
        aaaa={}
        ptr={}
        cnames={}
        #sanitize vmname
        vmname.gsub!('_','--')
        vmname.downcase!
        #check if vmname may be used in DNS
        if(vmname=~/^[a-z0-9-]+$/)
            if(!vmconfig.has_key?("vm"))
                fail("dns: VM #{vmname} does not have a hardware definition!")
            end
            if((!vmconfig['vm'].has_key?("network"))||(vmconfig['vm']['network'].empty?))
                fail("dns: VM #{vmname} does not have a network definition!")
            end
            netconfig=Hash[vmconfig['vm']['network']] #copy subhash network, there seems to be no simply way of just copying the whole hash...
            primary=netconfig.shift
            if(/^([a-fA-F0-9]{1,2}:){5}[a-fA-F0-9]{1,2}$/=~primary[1])
                addr=call_function("network::getipv6",primary[1],primary[0])
                aaaa.merge!({vmname=>[addr]})
                ptr.merge!({get_ptr(addr)=>"#{vmname}.#{primary_zone}."})
            else
                Puppet.warning("netinfo: node #{vmname}: invalid MAC address '#{primary[1]}'")
            end
            netconfig.each do |net,mac|
                if(/^([a-fA-F0-9]{1,2}:){5}[a-fA-F0-9]{1,2}$/=~mac)
                    addr=call_function("network::getipv6",mac,net)
                    aaaa.merge!({vmname+"-"+net=>[addr]})
                    ptr.merge!({get_ptr(addr)=>vmname+"-"+net+"."+primary_zone+"."})
                else
                    Puppet.warning("netinfo: node #{vmname}: invalid MAC address '#{mac}'")
                end
            end
            if(vmconfig.has_key?("cnames"))
                vmconfig['cnames'].each do|cname|
                    cnames.merge!({cname=>vmname})
                    #records+=get_dns_record('cname',cname,vmname)
                end
            end
        end
        {'AAAA'=>aaaa,'CNAME'=>cnames,'PTR'=>ptr}
    end

    def get_ptr(addr)
        out=""
        addr.tr(":","").reverse.split('').each do|c|
            out+=c+"."
        end
        out+"ip6.arpa."
    end
end
